@@ -51,11 +51,8 @@ import org.apache.sysml.runtime.matrix.MatrixCharacteristics;
 
 public class ReorgOp extends Hop implements MultiThreadedHop
 {
-	
 	public static boolean FORCE_DIST_SORT_INDEXES = false;
 	
-	public boolean bSortSPRewriteApplicable = false;
-	
 	private ReOrgOp op;
 	private int _maxNumThreads = -1; //-1 for unlimited
 	
@@ -229,29 +226,24 @@ public class ReorgOp extends Hop implements MultiThreadedHop
 			}
 			case RESHAPE:
 			{
+				Lop[] linputs = new Lop[4]; //main, rows, cols, byrow
+				for( int i=0; i<4; i++ )
+					linputs[i] = getInput().get(i).constructLops();
+				
 				if( et==ExecType.MR )
 				{
-					Transform transform1 = new Transform( getInput().get(0).constructLops(), 
-							HopsTransf2Lops.get(op), getDataType(), getValueType(), et);
+					Transform transform1 = new Transform( linputs,
+						HopsTransf2Lops.get(op), getDataType(), getValueType(), et);
 					setOutputDimensions(transform1);
 					setLineNumbers(transform1);
-					for( int i=1; i<=3; i++ ) //rows, cols, byrow
-					{
-						Lop ltmp = getInput().get(i).constructLops();
-						transform1.addInput(ltmp);
-						ltmp.addOutput(transform1);
-					}
-					transform1.setLevel(); //force order of added lops
 					
-					Group group1 = new Group(
-							transform1, Group.OperationTypes.Sort, DataType.MATRIX,
-							getValueType());
+					Group group1 = new Group(transform1,
+						Group.OperationTypes.Sort, DataType.MATRIX, getValueType());
 					setOutputDimensions(group1);
 					setLineNumbers(group1);
 	
-					Aggregate agg1 = new Aggregate(
-							group1, Aggregate.OperationTypes.Sum, DataType.MATRIX,
-							getValueType(), et);
+					Aggregate agg1 = new Aggregate(group1,
+						Aggregate.OperationTypes.Sum, DataType.MATRIX, getValueType(), et);
 					setOutputDimensions(agg1);
 					setLineNumbers(agg1);
 					
@@ -259,19 +251,11 @@ public class ReorgOp extends Hop implements MultiThreadedHop
 				}
 				else //CP/SPARK
 				{
-					Transform transform1 = new Transform( getInput().get(0).constructLops(), 
-							HopsTransf2Lops.get(op), getDataType(), getValueType(), et);
+					Transform transform1 = new Transform( linputs,
+						HopsTransf2Lops.get(op), getDataType(), getValueType(), et);
 					setOutputDimensions(transform1);
 					setLineNumbers(transform1);
 					
-					for( int i=1; i<=3; i++ ) //rows, cols, byrow
-					{
-						Lop ltmp = getInput().get(i).constructLops();
-						transform1.addInput(ltmp);
-						ltmp.addOutput(transform1);
-					}
-					transform1.setLevel(); //force order of added lops
-					
 					setLops(transform1);
 				}
 				break;
@@ -331,7 +315,7 @@ public class ReorgOp extends Hop implements MultiThreadedHop
 					else
 					{
 						//small vector, use in-memory sort
-						ArrayList<Hop> sinputs = new ArrayList<Hop>();
+						ArrayList<Hop> sinputs = new ArrayList<>();
 						sinputs.add(vinput);
 						sinputs.add(new LiteralOp(1)); //by (always vector)
 						sinputs.add(desc);
@@ -341,7 +325,7 @@ public class ReorgOp extends Hop implements MultiThreadedHop
 						//explicitly construct CP lop; otherwise there is danger of infinite recursion if forced runtime platform.
 						voutput.setLops( constructCPOrSparkSortLop(vinput, sinputs.get(1), sinputs.get(2), sinputs.get(3), ExecType.CP, false) );
 						voutput.getLops().getOutputParameters().setDimensions(vinput.getDim1(), vinput.getDim2(), vinput.getRowsInBlock(), vinput.getColsInBlock(), vinput.getNnz());
-						setLops( voutput.constructLops() );								
+						setLops( voutput.constructLops() );
 					}
 					
 					//Step 3: Data permutation (only required for sorting data) 
@@ -371,18 +355,22 @@ public class ReorgOp extends Hop implements MultiThreadedHop
 						setLops( mmult.constructLops() );
 						
 						//cleanups
-						HopRewriteUtils.removeChildReference(table, input);		
+						HopRewriteUtils.removeChildReference(table, input);
 					}
 				}
-				else //CP or Spark
+				else if( et==ExecType.SPARK ) {
+					boolean sortRewrite = !FORCE_DIST_SORT_INDEXES 
+						&& isSortSPRewriteApplicable() && by.getDataType().isScalar();
+					Lop transform1 = constructCPOrSparkSortLop(input, by, desc, ixret, et, sortRewrite);
+					setOutputDimensions(transform1);
+					setLineNumbers(transform1);
+					setLops(transform1);
+				}
+				else //CP
 				{
-					if( et==ExecType.SPARK && !FORCE_DIST_SORT_INDEXES)
-						bSortSPRewriteApplicable = isSortSPRewriteApplicable();
-					
-					Lop transform1 = constructCPOrSparkSortLop(input, by, desc, ixret, et, bSortSPRewriteApplicable);
+					Lop transform1 = constructCPOrSparkSortLop(input, by, desc, ixret, et, false);
 					setOutputDimensions(transform1);
 					setLineNumbers(transform1);
-					
 					setLops(transform1);
 				}
 				break;
@@ -401,23 +389,16 @@ public class ReorgOp extends Hop implements MultiThreadedHop
 	private static Lop constructCPOrSparkSortLop( Hop input, Hop by, Hop desc, Hop ixret, ExecType et, boolean bSortIndInMem ) 
 		throws HopsException, LopsException
 	{
-		Transform transform1 = new Transform( input.constructLops(), HopsTransf2Lops.get(ReOrgOp.SORT), 
-				     input.getDataType(), input.getValueType(), et, bSortIndInMem);
-		
-		for( Hop c : new Hop[]{by,desc,ixret} ) {
-			Lop ltmp = c.constructLops();
-			transform1.addInput(ltmp);
-			ltmp.addOutput(transform1);
-		}
-		
-		transform1.setLevel(); //force order of added lops
-		
-		return transform1;
+		Hop[] hinputs = new Hop[]{input, by, desc, ixret};
+		Lop[] linputs = new Lop[4];
+		for( int i=0; i<4; i++ )
+			linputs[i] = hinputs[i].constructLops();
+		return new Transform( linputs, HopsTransf2Lops.get(ReOrgOp.SORT), 
+			input.getDataType(), input.getValueType(), et, bSortIndInMem);
 	}
-			
+	
 	@Override
-	protected double computeOutputMemEstimate( long dim1, long dim2, long nnz )
-	{		
+	protected double computeOutputMemEstimate( long dim1, long dim2, long nnz ) {
 		//no dedicated mem estimation per op type, because always propagated via refreshSizeInformation
 		double sparsity = OptimizerUtils.getSparsity(dim1, dim2, nnz);
 		return OptimizerUtils.estimateSizeExactSparsity(dim1, dim2, sparsity);
