@@ -24,6 +24,8 @@ import java.util.HashMap;
 import org.apache.sysml.lops.Checkpoint;
 import org.apache.sysml.lops.Compression;
 import org.apache.sysml.lops.DataGen;
+import org.apache.sysml.lops.LeftIndex;
+import org.apache.sysml.lops.RightIndex;
 import org.apache.sysml.lops.WeightedCrossEntropy;
 import org.apache.sysml.lops.WeightedCrossEntropyR;
 import org.apache.sysml.lops.WeightedDivMM;
@@ -44,6 +46,7 @@ import org.apache.sysml.runtime.instructions.spark.AppendRSPInstruction;
 import org.apache.sysml.runtime.instructions.spark.ArithmeticBinarySPInstruction;
 import org.apache.sysml.runtime.instructions.spark.BinUaggChainSPInstruction;
 import org.apache.sysml.runtime.instructions.spark.BuiltinBinarySPInstruction;
+import org.apache.sysml.runtime.instructions.spark.BuiltinNarySPInstruction;
 import org.apache.sysml.runtime.instructions.spark.BuiltinUnarySPInstruction;
 import org.apache.sysml.runtime.instructions.spark.CSVReblockSPInstruction;
 import org.apache.sysml.runtime.instructions.spark.CastSPInstruction;
@@ -87,7 +90,7 @@ public class SPInstructionParser extends InstructionParser
 {	
 	public static final HashMap<String, SPINSTRUCTION_TYPE> String2SPInstructionType;
 	static {
-		String2SPInstructionType = new HashMap<String, SPInstruction.SPINSTRUCTION_TYPE>();
+		String2SPInstructionType = new HashMap<>();
 		
 		//unary aggregate operators
 		String2SPInstructionType.put( "uak+"   	, SPINSTRUCTION_TYPE.AggregateUnary);
@@ -141,9 +144,9 @@ public class SPInstructionParser extends InstructionParser
 		String2SPInstructionType.put( "maxpooling",             SPINSTRUCTION_TYPE.Convolution);
 		String2SPInstructionType.put( "relu_maxpooling",          SPINSTRUCTION_TYPE.Convolution);
 		
-		String2SPInstructionType.put( "rangeReIndex"   	, SPINSTRUCTION_TYPE.MatrixIndexing);
-		String2SPInstructionType.put( "leftIndex"   	, SPINSTRUCTION_TYPE.MatrixIndexing);
-		String2SPInstructionType.put( "mapLeftIndex"   	, SPINSTRUCTION_TYPE.MatrixIndexing);
+		String2SPInstructionType.put( RightIndex.OPCODE, SPINSTRUCTION_TYPE.MatrixIndexing);
+		String2SPInstructionType.put( LeftIndex.OPCODE, SPINSTRUCTION_TYPE.MatrixIndexing);
+		String2SPInstructionType.put( "mapLeftIndex" , SPINSTRUCTION_TYPE.MatrixIndexing);
 		
 		// Reorg Instruction Opcodes (repositioning of existing values)
 		String2SPInstructionType.put( "r'"   	   , SPINSTRUCTION_TYPE.Reorg);
@@ -214,6 +217,9 @@ public class SPInstructionParser extends InstructionParser
 		String2SPInstructionType.put( "asin"  , SPINSTRUCTION_TYPE.BuiltinUnary);
 		String2SPInstructionType.put( "acos"  , SPINSTRUCTION_TYPE.BuiltinUnary);
 		String2SPInstructionType.put( "atan"  , SPINSTRUCTION_TYPE.BuiltinUnary);
+		String2SPInstructionType.put( "sinh"   , SPINSTRUCTION_TYPE.BuiltinUnary);
+		String2SPInstructionType.put( "cosh"   , SPINSTRUCTION_TYPE.BuiltinUnary);
+		String2SPInstructionType.put( "tanh"   , SPINSTRUCTION_TYPE.BuiltinUnary);
 		String2SPInstructionType.put( "sign"  , SPINSTRUCTION_TYPE.BuiltinUnary);
 		String2SPInstructionType.put( "sqrt"  , SPINSTRUCTION_TYPE.BuiltinUnary);
 		String2SPInstructionType.put( "plogp" , SPINSTRUCTION_TYPE.BuiltinUnary);
@@ -238,6 +244,8 @@ public class SPInstructionParser extends InstructionParser
 		String2SPInstructionType.put( "rappend", SPINSTRUCTION_TYPE.RAppend);
 		String2SPInstructionType.put( "gappend", SPINSTRUCTION_TYPE.GAppend);
 		String2SPInstructionType.put( "galignedappend", SPINSTRUCTION_TYPE.GAlignedAppend);
+		String2SPInstructionType.put( "cbind", SPINSTRUCTION_TYPE.BuiltinNary);
+		String2SPInstructionType.put( "rbind", SPINSTRUCTION_TYPE.BuiltinNary);
 		
 		String2SPInstructionType.put( DataGen.RAND_OPCODE  , SPINSTRUCTION_TYPE.Rand);
 		String2SPInstructionType.put( DataGen.SEQ_OPCODE   , SPINSTRUCTION_TYPE.Rand);
@@ -394,7 +402,10 @@ public class SPInstructionParser extends InstructionParser
 				
 			case BuiltinUnary:
 				return BuiltinUnarySPInstruction.parseInstruction(str);
-				
+			
+			case BuiltinNary:
+				return BuiltinNarySPInstruction.parseInstruction(str);
+			
 			case ParameterizedBuiltin:
 				return ParameterizedBuiltinSPInstruction.parseInstruction(str);
 				
