@@ -24,8 +24,6 @@ import static jcuda.jcudnn.JCudnn.cudnnCreate;
 import static jcuda.jcudnn.JCudnn.cudnnDestroy;
 import static jcuda.jcusolver.JCusolverDn.cusolverDnCreate;
 import static jcuda.jcusolver.JCusolverDn.cusolverDnDestroy;
-import static jcuda.jcusolver.JCusolverSp.cusolverSpCreate;
-import static jcuda.jcusolver.JCusolverSp.cusolverSpDestroy;
 import static jcuda.jcusparse.JCusparse.cusparseCreate;
 import static jcuda.jcusparse.JCusparse.cusparseDestroy;
 import static jcuda.runtime.JCuda.cudaDeviceScheduleBlockingSync;
@@ -41,14 +39,18 @@ import java.util.ArrayList;
 import java.util.Collections;
 import java.util.Comparator;
 import java.util.HashMap;
-import java.util.LinkedList;
+import java.util.HashSet;
+import java.util.Iterator;
 import java.util.Map;
+import java.util.Map.Entry;
+import java.util.Set;
 
 import org.apache.commons.logging.Log;
 import org.apache.commons.logging.LogFactory;
 import org.apache.sysml.api.DMLScript;
 import org.apache.sysml.conf.ConfigurationManager;
 import org.apache.sysml.conf.DMLConfig;
+import org.apache.sysml.hops.OptimizerUtils;
 import org.apache.sysml.runtime.DMLRuntimeException;
 import org.apache.sysml.runtime.controlprogram.caching.MatrixObject;
 import org.apache.sysml.runtime.instructions.gpu.GPUInstruction;
@@ -59,7 +61,6 @@ import jcuda.Pointer;
 import jcuda.jcublas.cublasHandle;
 import jcuda.jcudnn.cudnnHandle;
 import jcuda.jcusolver.cusolverDnHandle;
-import jcuda.jcusolver.cusolverSpHandle;
 import jcuda.jcusparse.cusparseHandle;
 import jcuda.runtime.JCuda;
 import jcuda.runtime.cudaDeviceProp;
@@ -86,13 +87,34 @@ public class GPUContext {
 	 * active device assigned to this GPUContext instance
 	 */
 	private final int deviceNum;
+	/**
+	 * cudnnHandle for Deep Neural Network operations on the GPU
+	 */
+	private cudnnHandle cudnnHandle;
+	/**
+	 * cublasHandle for BLAS operations on the GPU
+	 */
+	private cublasHandle cublasHandle;
+	/**
+	 * cusparseHandle for certain sparse BLAS operations on the GPU
+	 */
+	private cusparseHandle cusparseHandle;
+	/**
+	 * cusolverDnHandle for invoking solve() function on dense matrices on the GPU
+	 */
+	private cusolverDnHandle cusolverDnHandle;
+	/**
+	 * to launch custom CUDA kernel, specific to the active GPU for this GPUContext
+	 */
+	private JCudaKernels kernels;
+
 	// Invoke cudaMemGetInfo to get available memory information. Useful if GPU is shared among multiple application.
 	public double GPU_MEMORY_UTILIZATION_FACTOR = ConfigurationManager.getDMLConfig()
 			.getDoubleValue(DMLConfig.GPU_MEMORY_UTILIZATION_FACTOR);
 	/**
 	 * Map of free blocks allocate on GPU. maps size_of_block -> pointer on GPU
 	 */
-	private LRUCacheMap<Long, LinkedList<Pointer>> freeCUDASpaceMap = new LRUCacheMap<>();
+	private LRUCacheMap<Long, Set<Pointer>> freeCUDASpaceMap = new LRUCacheMap<>();
 	/**
 	 * To record size of allocated blocks
 	 */
@@ -105,30 +127,6 @@ public class GPUContext {
 	 * so that an extraneous host to dev transfer can be avoided
 	 */
 	private ArrayList<GPUObject> allocatedGPUObjects = new ArrayList<>();
-	/**
-	 * cudnnHandle for Deep Neural Network operations on the GPU
-	 */
-	private final ThreadLocal<cudnnHandle> cudnnHandle = new ThreadLocal<>();
-	/**
-	 * cublasHandle for BLAS operations on the GPU
-	 */
-	private final ThreadLocal<cublasHandle> cublasHandle = new ThreadLocal<>();
-	/**
-	 * cusparseHandle for certain sparse BLAS operations on the GPU
-	 */
-	private final ThreadLocal<cusparseHandle> cusparseHandle = new ThreadLocal<>();
-	/**
-	 * cusolverDnHandle for invoking solve() function on dense matrices on the GPU
-	 */
-	private final ThreadLocal<cusolverDnHandle> cusolverDnHandle = new ThreadLocal<>();
-	/**
-	 * cusolverSpHandle for invoking solve() function on sparse matrices on the GPU
-	 */
-	private final ThreadLocal<cusolverSpHandle> cusolverSpHandle = new ThreadLocal<>();
-	/**
-	 * to launch custom CUDA kernel, specific to the active GPU for this GPUContext
-	 */
-	private final ThreadLocal<JCudaKernels> kernels = new ThreadLocal<>();
 
 	protected GPUContext(int deviceNum) throws DMLRuntimeException {
 		this.deviceNum = deviceNum;
@@ -151,45 +149,16 @@ public class GPUContext {
 		LOG.info(" GPU memory - Total: " + (total[0] * (1e-6)) + " MB, Available: " + (free[0] * (1e-6)) + " MB on "
 				+ this);
 
-	}
-
-	private void initializeCudaLibraryHandles() throws DMLRuntimeException {
-		if (cudnnHandle.get() == null) {
-			cudnnHandle.set(new cudnnHandle());
-			cudnnCreate(cudnnHandle.get());
-		}
-
-		if (cublasHandle.get() == null) {
-			cublasHandle.set(new cublasHandle());
-			cublasCreate(cublasHandle.get());
-		}
-		// For cublas v2, cublasSetPointerMode tells Cublas whether to expect scalar arguments on device or on host
-		// This applies to arguments like "alpha" in Dgemm, and "y" in Ddot.
-		// cublasSetPointerMode(LibMatrixCUDA.cublasHandle, cublasPointerMode.CUBLAS_POINTER_MODE_DEVICE);
-
-		if (cusparseHandle.get() == null) {
-			cusparseHandle.set(new cusparseHandle());
-			cusparseCreate(cusparseHandle.get());
-		}
-
-		if (cusolverDnHandle.get() == null) {
-			cusolverDnHandle.set(new cusolverDnHandle());
-			cusolverDnCreate(cusolverDnHandle.get());
-		}
-
-		if (cusolverSpHandle.get() == null) {
-			cusolverSpHandle.set(new cusolverSpHandle());
-			cusolverSpCreate(cusolverSpHandle.get());
-		}
-
-		if (kernels.get() == null) {
-			kernels.set(new JCudaKernels());
+		if (GPUContextPool.initialGPUMemBudget() > OptimizerUtils.getLocalMemBudget()) {
+			LOG.warn("Potential under-utilization: GPU memory (" + GPUContextPool.initialGPUMemBudget()
+					+ ") > driver memory budget (" + OptimizerUtils.getLocalMemBudget() + "). "
+					+ "Consider increasing the driver memory budget.");
 		}
 	}
 
 	/**
 	 * Returns which device is currently being used.
-	 * 
+	 *
 	 * @return the current device for the calling host thread
 	 */
 	public static int cudaGetDevice() {
@@ -198,9 +167,74 @@ public class GPUContext {
 		return device[0];
 	}
 
+	/**
+	 * Print information of memory usage.
+	 *
+	 * @param opcode opcode of caller
+	 * @throws DMLRuntimeException if error
+	 */
+	public void printMemoryInfo(String opcode) throws DMLRuntimeException {
+		if (LOG.isDebugEnabled()) {
+			long totalFreeCUDASpace = 0;
+			for (Entry<Long, Set<Pointer>> kv : freeCUDASpaceMap.entrySet()) {
+				totalFreeCUDASpace += kv.getKey() * kv.getValue().size();
+			}
+			long readLockedAllocatedMemory = 0;
+			long writeLockedAllocatedMemory = 0;
+			long unlockedAllocatedMemory = 0;
+			for (GPUObject gpuObj : allocatedGPUObjects) {
+				if (gpuObj.readLocks.longValue() > 0)
+					readLockedAllocatedMemory += gpuObj.getSizeOnDevice();
+				else if (gpuObj.writeLock)
+					writeLockedAllocatedMemory += gpuObj.getSizeOnDevice();
+				else
+					unlockedAllocatedMemory += gpuObj.getSizeOnDevice();
+			}
+			long free[] = { 0 };
+			long total[] = { 0 };
+			cudaMemGetInfo(free, total);
+			long gpuFreeMemory = (long) (free[0] * GPU_MEMORY_UTILIZATION_FACTOR);
+			LOG.debug(opcode + ": Total memory: " + total[0] + ", Free memory: " + free[0] + " (with util factor: "
+					+ gpuFreeMemory + "), " + "Lazy unfreed memory: " + totalFreeCUDASpace
+					+ ", Locked allocated memory (read/write): " + readLockedAllocatedMemory + "/"
+					+ writeLockedAllocatedMemory + ", " + " Unlocked allocated memory: " + unlockedAllocatedMemory);
+		}
+	}
+
+	private void initializeCudaLibraryHandles() throws DMLRuntimeException {
+		deleteCudaLibraryHandles();
+
+		if (cudnnHandle == null) {
+			cudnnHandle = new cudnnHandle();
+			cudnnCreate(cudnnHandle);
+		}
+
+		if (cublasHandle == null) {
+			cublasHandle = new cublasHandle();
+			cublasCreate(cublasHandle);
+		}
+		// For cublas v2, cublasSetPointerMode tells Cublas whether to expect scalar arguments on device or on host
+		// This applies to arguments like "alpha" in Dgemm, and "y" in Ddot.
+		// cublasSetPointerMode(LibMatrixCUDA.cublasHandle, cublasPointerMode.CUBLAS_POINTER_MODE_DEVICE);
+
+		if (cusparseHandle == null) {
+			cusparseHandle = new cusparseHandle();
+			cusparseCreate(cusparseHandle);
+		}
+
+		if (cusolverDnHandle == null) {
+			cusolverDnHandle = new cusolverDnHandle();
+			cusolverDnCreate(cusolverDnHandle);
+		}
+		
+		if (kernels == null) {
+			kernels = new JCudaKernels();
+		}
+	}
+
 	/**
 	 * Returns which device is assigned to this GPUContext instance.
-	 * 
+	 *
 	 * @return active device assigned to this GPUContext instance
 	 */
 	public int getDeviceNum() {
@@ -213,7 +247,7 @@ public class GPUContext {
 	 * {@link org.apache.sysml.runtime.controlprogram.context.ExecutionContext#getGPUContext(int)}
 	 * If in a multi-threaded environment like parfor, this method must be called when in the
 	 * appropriate thread.
-	 * 
+	 *
 	 * @throws DMLRuntimeException if DMLRuntimeException occurs
 	 */
 	public void initializeThread() throws DMLRuntimeException {
@@ -259,21 +293,30 @@ public class GPUContext {
 		long t0 = 0, t1 = 0, end = 0;
 		Pointer A;
 		if (freeCUDASpaceMap.containsKey(size)) {
-			LOG.trace("GPU : in allocate from instruction " + instructionName + ", found free block of size " + (size
-					/ 1024.0) + " Kbytes from previously allocated block on " + this);
-			if (instructionName != null && GPUStatistics.DISPLAY_STATISTICS)
+			if (LOG.isTraceEnabled()) {
+				LOG.trace(
+						"GPU : in allocate from instruction " + instructionName + ", found free block of size " + (size
+								/ 1024.0) + " Kbytes from previously allocated block on " + this);
+			}
+			if (instructionName != null && DMLScript.FINEGRAINED_STATISTICS)
 				t0 = System.nanoTime();
-			LinkedList<Pointer> freeList = freeCUDASpaceMap.get(size);
-			A = freeList.pop();
+			Set<Pointer> freeList = freeCUDASpaceMap.get(size);
+
+			Iterator<Pointer> it = freeList.iterator(); // at this point, freeList should have at least one element
+			A = it.next();
+			it.remove();
+
 			if (freeList.isEmpty())
 				freeCUDASpaceMap.remove(size);
-			if (instructionName != null && GPUStatistics.DISPLAY_STATISTICS)
+			if (instructionName != null && DMLScript.FINEGRAINED_STATISTICS)
 				GPUStatistics
-				.maintainCPMiscTimes(instructionName, GPUInstruction.MISC_TIMER_REUSE, System.nanoTime() - t0);
+						.maintainCPMiscTimes(instructionName, GPUInstruction.MISC_TIMER_REUSE, System.nanoTime() - t0);
 		} else {
-			LOG.trace(
-					"GPU : in allocate from instruction " + instructionName + ", allocating new block of size " + (size
-							/ 1024.0) + " Kbytes on " + this);
+			if (LOG.isTraceEnabled()) {
+				LOG.trace(
+						"GPU : in allocate from instruction " + instructionName + ", allocating new block of size " + (
+								size / 1024.0) + " Kbytes on " + this);
+			}
 			if (DMLScript.STATISTICS)
 				t0 = System.nanoTime();
 			ensureFreeSpace(instructionName, size);
@@ -283,19 +326,21 @@ public class GPUContext {
 				GPUStatistics.cudaAllocTime.add(System.nanoTime() - t0);
 			if (DMLScript.STATISTICS)
 				GPUStatistics.cudaAllocCount.add(statsCount);
-			if (instructionName != null && GPUStatistics.DISPLAY_STATISTICS)
+			if (instructionName != null && DMLScript.FINEGRAINED_STATISTICS)
 				GPUStatistics.maintainCPMiscTimes(instructionName, GPUInstruction.MISC_TIMER_ALLOCATE,
 						System.nanoTime() - t0);
 		}
 		// Set all elements to 0 since newly allocated space will contain garbage
 		if (DMLScript.STATISTICS)
 			t1 = System.nanoTime();
-		LOG.trace("GPU : in allocate from instruction " + instructionName + ", setting block of size " + (size / 1024.0)
-				+ " Kbytes to zero on " + this);
+		if (LOG.isTraceEnabled()) {
+			LOG.trace("GPU : in allocate from instruction " + instructionName + ", setting block of size " + (size
+					/ 1024.0) + " Kbytes to zero on " + this);
+		}
 		cudaMemset(A, 0, size);
 		if (DMLScript.STATISTICS)
 			end = System.nanoTime();
-		if (instructionName != null && GPUStatistics.DISPLAY_STATISTICS)
+		if (instructionName != null && DMLScript.FINEGRAINED_STATISTICS)
 			GPUStatistics.maintainCPMiscTimes(instructionName, GPUInstruction.MISC_TIMER_SET_ZERO, end - t1);
 		if (DMLScript.STATISTICS)
 			GPUStatistics.cudaMemSet0Time.add(end - t1);
@@ -312,7 +357,7 @@ public class GPUContext {
 	 * @param toFree {@link Pointer} instance to be freed
 	 */
 	public void cudaFreeHelper(final Pointer toFree) {
-		cudaFreeHelper(null, toFree, false);
+		cudaFreeHelper(null, toFree, DMLScript.EAGER_CUDA_FREE);
 	}
 
 	/**
@@ -332,7 +377,7 @@ public class GPUContext {
 	 * @param toFree          {@link Pointer} instance to be freed
 	 */
 	public void cudaFreeHelper(String instructionName, final Pointer toFree) {
-		cudaFreeHelper(instructionName, toFree, false);
+		cudaFreeHelper(instructionName, toFree, DMLScript.EAGER_CUDA_FREE);
 	}
 
 	/**
@@ -344,37 +389,46 @@ public class GPUContext {
 	 */
 	public void cudaFreeHelper(String instructionName, final Pointer toFree, boolean eager) {
 		Pointer dummy = new Pointer();
-		if (toFree == dummy) // trying to free a null pointer
+		if (toFree == dummy) { // trying to free a null pointer
+			if (LOG.isTraceEnabled()) {
+				LOG.trace("GPU : trying to free an empty pointer");
+			}
 			return;
+		}
 		long t0 = 0;
-		assert cudaBlockSizeMap.containsKey(
-				toFree) : "ERROR : Internal state corrupted, cache block size map is not aware of a block it trying to free up";
-				long size = cudaBlockSizeMap.get(toFree);
-				if (eager) {
-					LOG.trace("GPU : eagerly freeing cuda memory [ " + toFree + " ] for instruction " + instructionName + " on "
-							+ this);
-					if (DMLScript.STATISTICS)
-						t0 = System.nanoTime();
-					cudaFree(toFree);
-					cudaBlockSizeMap.remove(toFree);
-					if (DMLScript.STATISTICS)
-						GPUStatistics.cudaDeAllocTime.add(System.nanoTime() - t0);
-					if (DMLScript.STATISTICS)
-						GPUStatistics.cudaDeAllocCount.add(1);
-					if (instructionName != null && GPUStatistics.DISPLAY_STATISTICS)
-						GPUStatistics.maintainCPMiscTimes(instructionName, GPUInstruction.MISC_TIMER_CUDA_FREE,
-								System.nanoTime() - t0);
-				} else {
-					LOG.trace("GPU : lazily freeing cuda memory for instruction " + instructionName + " on " + this);
-					LinkedList<Pointer> freeList = freeCUDASpaceMap.get(size);
-					if (freeList == null) {
-						freeList = new LinkedList<Pointer>();
-						freeCUDASpaceMap.put(size, freeList);
-					}
-					if (freeList.contains(toFree))
-						throw new RuntimeException("GPU : Internal state corrupted, double free");
-					freeList.add(toFree);
-				}
+		if (!cudaBlockSizeMap.containsKey(toFree))
+			throw new RuntimeException(
+					"ERROR : Internal state corrupted, cache block size map is not aware of a block it trying to free up");
+		long size = cudaBlockSizeMap.get(toFree);
+		if (eager) {
+			if (LOG.isTraceEnabled()) {
+				LOG.trace("GPU : eagerly freeing cuda memory [ " + toFree + " ] of size " + size + " for instruction " + instructionName
+						+ " on " + this);
+			}
+			if (DMLScript.STATISTICS)
+				t0 = System.nanoTime();
+			cudaFree(toFree);
+			cudaBlockSizeMap.remove(toFree);
+			if (DMLScript.STATISTICS)
+				GPUStatistics.cudaDeAllocTime.add(System.nanoTime() - t0);
+			if (DMLScript.STATISTICS)
+				GPUStatistics.cudaDeAllocCount.add(1);
+			if (instructionName != null && DMLScript.FINEGRAINED_STATISTICS)
+				GPUStatistics.maintainCPMiscTimes(instructionName, GPUInstruction.MISC_TIMER_CUDA_FREE,
+						System.nanoTime() - t0);
+		} else {
+			if (LOG.isTraceEnabled()) {
+				LOG.trace("GPU : lazily freeing cuda memory of size " + size + " for instruction " + instructionName + " on " + this);
+			}
+			Set<Pointer> freeList = freeCUDASpaceMap.get(size);
+			if (freeList == null) {
+				freeList = new HashSet<>();
+				freeCUDASpaceMap.put(size, freeList);
+			}
+			if (freeList.contains(toFree))
+				throw new RuntimeException("GPU : Internal state corrupted, double free");
+			freeList.add(toFree);
+		}
 	}
 
 	/**
@@ -395,9 +449,10 @@ public class GPUContext {
 	 * @throws DMLRuntimeException if DMLRuntimeException occurs
 	 */
 	void ensureFreeSpace(String instructionName, long size) throws DMLRuntimeException {
-		if (size >= getAvailableMemory()) {
+		if (size < 0)
+			throw new DMLRuntimeException("The size cannot be negative:" + size);
+		else if (size >= getAvailableMemory())
 			evict(instructionName, size);
-		}
 	}
 
 	/**
@@ -425,18 +480,28 @@ public class GPUContext {
 	 * @throws DMLRuntimeException If no reusable memory blocks to free up or if not enough matrix blocks with zero locks on them.
 	 */
 	protected void evict(String instructionName, final long neededSize) throws DMLRuntimeException {
-		LOG.trace("GPU : evict called from " + instructionName + " for size " + neededSize + " on " + this);
+		if (LOG.isTraceEnabled()) {
+			LOG.trace("GPU : evict called from " + instructionName + " for size " + neededSize + " on " + this);
+		}
 		GPUStatistics.cudaEvictionCount.add(1);
+		if (LOG.isDebugEnabled()) {
+			printMemoryInfo("EVICTION_CUDA_FREE_SPACE");
+		}
+		
 		// Release the set of free blocks maintained in a GPUObject.freeCUDASpaceMap
 		// to free up space
-		LRUCacheMap<Long, LinkedList<Pointer>> lruCacheMap = freeCUDASpaceMap;
+		LRUCacheMap<Long, Set<Pointer>> lruCacheMap = freeCUDASpaceMap;
 		while (lruCacheMap.size() > 0) {
 			if (neededSize <= getAvailableMemory())
 				break;
-			Map.Entry<Long, LinkedList<Pointer>> toFreeListPair = lruCacheMap.removeAndGetLRUEntry();
-			LinkedList<Pointer> toFreeList = toFreeListPair.getValue();
+			Map.Entry<Long, Set<Pointer>> toFreeListPair = lruCacheMap.removeAndGetLRUEntry();
+			Set<Pointer> toFreeList = toFreeListPair.getValue();
 			Long size = toFreeListPair.getKey();
-			Pointer toFree = toFreeList.pop();
+
+			Iterator<Pointer> it = toFreeList.iterator(); // at this point, freeList should have at least one element
+			Pointer toFree = it.next();
+			it.remove();
+
 			if (toFreeList.isEmpty())
 				lruCacheMap.remove(size);
 			cudaFreeHelper(instructionName, toFree, true);
@@ -453,18 +518,19 @@ public class GPUContext {
 		Collections.sort(allocatedGPUObjects, new Comparator<GPUObject>() {
 			@Override
 			public int compare(GPUObject p1, GPUObject p2) {
-				long p1Val = p1.locks.get();
-				long p2Val = p2.locks.get();
-
-				if (p1Val > 0 && p2Val > 0) {
+				if (p1.isLocked() && p2.isLocked()) {
 					// Both are locked, so don't sort
 					return 0;
-				} else if (p1Val > 0 || p2Val > 0) {
+				} else if (p1.isLocked()) {
 					// Put the unlocked one to RHS
-					return Long.compare(p2Val, p1Val);
+					// a value less than 0 if x < y; and a value greater than 0 if x > y
+					return -1;
+				} else if (p2.isLocked()) {
+					// Put the unlocked one to RHS
+					// a value less than 0 if x < y; and a value greater than 0 if x > y
+					return 1;
 				} else {
 					// Both are unlocked
-
 					if (evictionPolicy == EvictionPolicy.MIN_EVICT) {
 						long p1Size = 0;
 						long p2Size = 0;
@@ -490,15 +556,18 @@ public class GPUContext {
 		});
 
 		while (neededSize > getAvailableMemory() && allocatedGPUObjects.size() > 0) {
+			if (LOG.isDebugEnabled()) {
+				printMemoryInfo("EVICTION_UNLOCKED");
+			}
 			GPUObject toBeRemoved = allocatedGPUObjects.get(allocatedGPUObjects.size() - 1);
-			if (toBeRemoved.locks.get() > 0) {
+			if (toBeRemoved.isLocked()) {
 				throw new DMLRuntimeException(
-						"There is not enough memory on device for this matrix, request (" + neededSize + ")");
+						"There is not enough memory on device for this matrix, request (" + neededSize
+								+ "). Allocated GPU objects:" + allocatedGPUObjects.toString());
 			}
 			if (toBeRemoved.dirty) {
-				toBeRemoved.copyFromDeviceToHost();
+				toBeRemoved.copyFromDeviceToHost(instructionName, true);
 			}
-
 			toBeRemoved.clearData(true);
 		}
 	}
@@ -574,7 +643,7 @@ public class GPUContext {
 
 	/**
 	 * Instantiates a new {@link GPUObject} initialized with the given {@link org.apache.sysml.runtime.controlprogram.caching.MatrixObject MatrixObject}.
-	 * 
+	 *
 	 * @param mo a {@link org.apache.sysml.runtime.controlprogram.caching.MatrixObject MatrixObject} that represents a matrix
 	 * @return a new {@link GPUObject} instance
 	 */
@@ -638,56 +707,47 @@ public class GPUContext {
 
 	/**
 	 * Returns the cudnnHandle for Deep Neural Network operations on the GPU.
-	 * 
+	 *
 	 * @return cudnnHandle for current thread
 	 */
 	public cudnnHandle getCudnnHandle() {
-		return cudnnHandle.get();
+		return cudnnHandle;
 	}
 
 	/**
 	 * Returns cublasHandle for BLAS operations on the GPU.
-	 * 
+	 *
 	 * @return cublasHandle for current thread
 	 */
 	public cublasHandle getCublasHandle() {
-		return cublasHandle.get();
+		return cublasHandle;
 	}
 
 	/**
 	 * Returns cusparseHandle for certain sparse BLAS operations on the GPU.
-	 * 
+	 *
 	 * @return cusparseHandle for current thread
 	 */
 	public cusparseHandle getCusparseHandle() {
-		return cusparseHandle.get();
+		return cusparseHandle;
 	}
 
 	/**
 	 * Returns cusolverDnHandle for invoking solve() function on dense matrices on the GPU.
-	 * 
+	 *
 	 * @return cusolverDnHandle for current thread
 	 */
 	public cusolverDnHandle getCusolverDnHandle() {
-		return cusolverDnHandle.get();
-	}
-
-	/**
-	 * Returns cusolverSpHandle for invoking solve() function on sparse matrices on the GPU.
-	 * 
-	 * @return cusolverSpHandle for current thread
-	 */
-	public cusolverSpHandle getCusolverSpHandle() {
-		return cusolverSpHandle.get();
+		return cusolverDnHandle;
 	}
 
 	/**
 	 * Returns utility class used to launch custom CUDA kernel, specific to the active GPU for this GPUContext.
-	 * 
+	 *
 	 * @return {@link JCudaKernels} for current thread
 	 */
 	public JCudaKernels getKernels() {
-		return kernels.get();
+		return kernels;
 	}
 
 	/**
@@ -696,13 +756,34 @@ public class GPUContext {
 	 * @throws DMLRuntimeException if error
 	 */
 	public void destroy() throws DMLRuntimeException {
-		LOG.trace("GPU : this context was destroyed, this = " + this.toString());
+		if (LOG.isTraceEnabled()) {
+			LOG.trace("GPU : this context was destroyed, this = " + this.toString());
+		}
 		clearMemory();
-		cudnnDestroy(cudnnHandle.get());
-		cublasDestroy(cublasHandle.get());
-		cusparseDestroy(cusparseHandle.get());
-		cusolverDnDestroy(cusolverDnHandle.get());
-		cusolverSpDestroy(cusolverSpHandle.get());
+
+		deleteCudaLibraryHandles();
+	}
+
+	/**
+	 *	Deletes CUDA library handles
+	 */
+	private void deleteCudaLibraryHandles() {
+		if (cudnnHandle != null)
+			cudnnDestroy(cudnnHandle);
+
+		if (cublasHandle != null)
+			cublasDestroy(cublasHandle);
+
+		if (cusparseHandle != null)
+			cusparseDestroy(cusparseHandle);
+
+		if (cusolverDnHandle != null)
+			cusolverDnDestroy(cusolverDnHandle);
+
+		cudnnHandle = null;
+		cublasHandle = null;
+		cusparseHandle = null;
+		cusolverDnHandle = null;
 	}
 
 	/**
@@ -721,7 +802,7 @@ public class GPUContext {
 			if (o.isDirty()) {
 				LOG.warn("Attempted to free GPU Memory when a block[" + o
 						+ "] is still on GPU memory, copying it back to host.");
-				o.acquireHostRead();
+				o.acquireHostRead(null);
 			}
 			o.clearData(true);
 		}
@@ -760,7 +841,7 @@ public class GPUContext {
 		}
 
 		// garbage collect all temporarily allocated spaces
-		for (LinkedList<Pointer> l : freeCUDASpaceMap.values()) {
+		for (Set<Pointer> l : freeCUDASpaceMap.values()) {
 			for (Pointer p : l) {
 				cudaFreeHelper(p, true);
 			}
